# noinspection PyPackageRequirements
from sanic.exceptions import NotFound

from ..api import Gino as _Gino
from ..local import enable_task_local, disable_task_local


class SanicModelMixin:
    @classmethod
    async def get_or_404(cls, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        rv = await cls.get(*args, **kwargs)
        if rv is None:
            raise NotFound('{} is not found'.format(cls.__name__))
        return rv


# noinspection PyClassHasNoInit
class Gino(_Gino):
    """Support Sanic web server.

    By :meth:`init_app` GINO registers a few hooks on Sanic, so that GINO could
    use database configuration in Sanic `config` to initialize the bound pool.

    A lazy connection context is enabled by default for every request. You can
    change this default behavior by setting `DB_USE_CONNECTION_FOR_REQUEST`
    config value to `False`. By default, a database connection is borrowed on
    the first query, shared in the same execution context, and returned to the
    pool on response. If you need to release the connection early in the middle
    to do some long-running tasks, you can simply do this:

        await request['connection'].release()

    Here `request['connection']` is a :class:`LazyConnection` object, see its
    doc string for more information.
    """
    default_model_classes = _Gino.default_model_classes + (SanicModelMixin,)

    def init_app(self, app):
        task_local_enabled = [False]

        if app.config.setdefault('DB_USE_CONNECTION_FOR_REQUEST', True):
            @app.middleware('request')
            async def on_request(request):
                request['connection_ctx'] = ctx = self.acquire(lazy=True)
                request['connection'] = await ctx.__aenter__()

            @app.middleware('response')
            async def on_response(request, _):
                ctx = request.pop('connection_ctx', None)
                request.pop('connection', None)
                if ctx is not None:
                    await ctx.__aexit__(None, None, None)

        @app.listener('before_server_start')
        async def before_server_start(_, loop):
            if app.config.setdefault('DB_USE_CONNECTION_FOR_REQUEST', True):
                enable_task_local(loop)
                task_local_enabled[0] = True

            await self.create_pool(
                host=app.config.setdefault('DB_HOST', 'localhost'),
                port=app.config.setdefault('DB_PORT', 5432),
                user=app.config.setdefault('DB_USER', 'postgres'),
                password=app.config.setdefault('DB_PASSWORD', ''),
                database=app.config.setdefault('DB_DATABASE', 'postgres'),
                min_size=app.config.setdefault('DB_POOL_MIN_SIZE', 5),
                max_size=app.config.setdefault('DB_POOL_MAX_SIZE', 10),
                loop=loop,
            )

        @app.listener('after_server_stop')
        async def after_server_stop(_, loop):
            await self._bind.close()
            if task_local_enabled[0]:
                disable_task_local(loop)
                task_local_enabled[0] = False



import random
import string

import asyncpg
import pytest
import sqlalchemy

from .models import db, DB_ARGS


@pytest.fixture(scope='module')
def engine():
    rv = sqlalchemy.create_engine(
        'postgresql://{user}:{password}@{host}:{port}/{database}'.format(
            **DB_ARGS))
    db.create_all(rv)
    yield rv
    db.drop_all(rv)
    rv.dispose()


# noinspection PyUnusedLocal,PyShadowingNames
@pytest.fixture
async def pool(engine):
    async with db.create_pool(**DB_ARGS) as rv:
        yield rv
        await rv.execute('DELETE FROM gino_users')


# noinspection PyUnusedLocal,PyShadowingNames
@pytest.fixture
async def asyncpg_pool(engine):
    async with asyncpg.create_pool(**DB_ARGS) as rv:
        yield rv
        await rv.execute('DELETE FROM gino_users')


@pytest.fixture
def random_name(length=8) -> str:
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))



import random

import pytest
from .models import db, DB_ARGS, User

pytestmark = pytest.mark.asyncio


# noinspection PyUnusedLocal
async def test_create(pool):
    nickname = 'test_create_{}'.format(random.random())
    u = await User.create(nickname=nickname)
    assert u.id is not None
    assert u.nickname == nickname
    return u


async def test_get(pool):
    u1 = await test_create(pool)
    u2 = await User.get(u1.id)
    assert u1.id == u2.id
    assert u1.nickname == u2.nickname
    assert u1 is not u2


async def test_unbind():
    try:
        async with db.create_pool(**DB_ARGS):
            await test_create(None)
        with pytest.raises(AttributeError):
            await test_create(None)
    finally:
        async with db.create_pool(**DB_ARGS) as pool:
            await pool.execute('DELETE FROM gino_users')



import sqlalchemy as sa


class ColumnAttribute:
    def __init__(self, column):
        self.name = column.name
        self.column = column

    def __get__(self, instance, owner):
        if instance is None:
            return self.column
        else:
            return instance.__values__.get(self.name)

    def __set__(self, instance, value):
        if instance is None:
            raise AttributeError('Cannot change columns.')
        else:
            instance.__values__[self.name] = value

    def __delete__(self, instance):
        raise AttributeError('Cannot delete column or value.')


class ModelType(type):
    def __iter__(self):
        # noinspection PyUnresolvedReferences
        return iter(self.__table__.columns)


class Model:
    __metadata__ = None
    __table__ = None
    __attr_factory__ = ColumnAttribute

    def __init__(self):
        self.__values__ = {}

    def __init_subclass__(cls, **kwargs):
        if cls.__table__ is None:
            cls.__table__ = cls._init_table(cls)

    @classmethod
    def _init_table(cls, sub_cls):
        table_name = getattr(sub_cls, '__tablename__', None)
        if table_name is None:
            return

        columns = []
        updates = {}
        for k, v in sub_cls.__dict__.items():
            if isinstance(v, sa.Column):
                v.name = k
                columns.append(v)
                updates[k] = sub_cls.__attr_factory__(v)

        # handle __table_args__
        table_args = getattr(sub_cls, '__table_args__', None)
        args, table_kw = (), {}
        if isinstance(table_args, dict):
            table_kw = table_args
        elif isinstance(table_args, tuple) and table_args:
            if isinstance(table_args[-1], dict):
                args, table_kw = table_args[0:-1], table_args[-1]
            else:
                args = table_args

        rv = sa.Table(table_name, sub_cls.__metadata__,
                      *columns, *args, **table_kw)
        for k, v in updates.items():
            setattr(sub_cls, k, v)
        return rv


def declarative_base(metadata, model_classes=(Model,), name='Model'):
    return ModelType(name, model_classes, {'__metadata__': metadata})


__all__ = ['ColumnAttribute', 'Model', 'declarative_base']



from .api import Gino
from .pool import GinoPool
from .connection import GinoConnection
from .local import get_local, enable_task_local, disable_task_local
from .exceptions import *

__version__ = '0.5.0'



cimport cpython

from asyncpg import Record
from asyncpg.protocol.record cimport ApgRecord_SET_ITEM


def update_record(record, i, val):
    assert isinstance(record, Record), 'asyncpg.Record is required'
    cpython.Py_DECREF(record[i])
    cpython.Py_INCREF(val)
    ApgRecord_SET_ITEM(record, i, val)



import asyncio
import functools
import types
from collections import deque

# noinspection PyProtectedMember
from asyncpg.connection import _ConnectionProxy
from asyncpg.exceptions import InterfaceError
from asyncpg.pool import Pool

from .local import get_local


class LazyConnection(_ConnectionProxy):
    """
    Use :class:`LazyConnection` to create a lazy connection which does not 
    immediately return a connection on creation. Connection is acquired before
    any public async method. User should explicitly call :meth:`get_connection`
    to get a real connection before calling public sync methods. And
    :meth:`release` should be called when the real connection is no longer 
    needed. Both methods can be called multiple times, and are coroutine-safe.
    """

    __slots__ = ('_pool', '_timeout', '_root', '_conn_cls', '_conn', '_closed',
                 '_conn_task', '_metadata', '_execution_options', '_loop',
                 '_wrap_cache')

    def __init__(self, pool, timeout, root, conn_cls, loop):
        self._pool = pool
        self._timeout = timeout
        self._root = root or self
        self._conn_cls = conn_cls
        self._conn = None
        self._conn_task = None
        self._metadata = pool.metadata
        self._execution_options = dict(pool.execution_options)
        self._loop = loop
        self._closed = False
        self._wrap_cache = {}

    @property
    def metadata(self):
        return self._metadata

    @property
    def execution_options(self):
        return self._execution_options

    async def get_connection(self):
        if self._closed:
            raise InterfaceError('This LazyConnection is returned to pool')
        if self._root is self:
            if self._conn_task is None:
                sup = super(self._pool.__class__, self._pool)
                self._conn_task = self._loop.create_task(
                    getattr(sup, '_acquire')(timeout=self._timeout))
            self._conn = await self._conn_task
            getattr(self._conn, '_set_metadata')(self._metadata)
        elif self._conn is None:
            self._conn = await self._root.get_connection()
        return self._conn

    async def release(self, close=False):
        self._closed = self._closed or close
        rv = None
        try:
            conn_to_release = None
            if self._conn_task is not None:
                fut, self._conn_task = self._conn_task, None
                conn_to_release = await fut
            if self._conn is not None:
                getattr(self._conn, '_set_metadata')(None)
            if conn_to_release is not None:
                sup = super(self._pool.__class__, self._pool)
                rv = await getattr(sup, 'release')(conn_to_release)
        finally:
            self._conn = None
        return rv

    def __getattr__(self, attr):
        rv = self._wrap_cache.get(attr)
        if rv is not None:
            # noinspection PyArgumentList
            return types.MethodType(rv, self)
        cls_val = getattr(self._conn_cls, attr)
        if asyncio.iscoroutinefunction(cls_val):
            @functools.wraps(cls_val)
            async def wrapper(_, *args, **kwargs):
                conn = await self.get_connection()
                return await getattr(conn, attr)(*args, **kwargs)
            self._wrap_cache[attr] = wrapper
            # noinspection PyArgumentList
            return types.MethodType(wrapper, self)
        if self._conn is None:
            raise InterfaceError(
                'Connection is not ready yet, or has been released')
        return getattr(self._conn, attr)


class GinoAcquireContext:
    __slots__ = ('timeout', 'connection', 'done', 'pool', '_reuse', '_lazy')

    def __init__(self, pool, timeout, reuse, lazy):
        self.pool = pool
        self.timeout = timeout
        self.connection = None
        self.done = False
        self._reuse = reuse
        self._lazy = lazy

    async def __aenter__(self):
        if self.connection is not None or self.done:
            raise InterfaceError('a connection is already acquired')
        self.connection = await getattr(self.pool, '_acquire')(
            self.timeout, reuse=self._reuse, lazy=self._lazy)
        return self.connection

    async def __aexit__(self, *exc):
        self.done = True
        con = self.connection
        self.connection = None
        await self.pool.release(con)

    def __await__(self):
        self.done = True
        return getattr(self.pool, '_acquire')(
            self.timeout, reuse=self._reuse, lazy=self._lazy).__await__()


class GinoPool(Pool):
    __slots__ = ('_metadata', '_execution_options')

    def __init__(self, metadata, *connect_args, min_size, max_size,
                 max_queries, max_inactive_connection_lifetime, setup, init,
                 loop, connection_class, **connect_kwargs):
        self._metadata = metadata
        self._execution_options = {}

        super().__init__(
            *connect_args, min_size=min_size, max_size=max_size,
            max_queries=max_queries,
            max_inactive_connection_lifetime=max_inactive_connection_lifetime,
            setup=setup, init=init, loop=loop,
            connection_class=connection_class, **connect_kwargs)

    async def _async__init__(self):
        rv = await super()._async__init__()
        if getattr(self._metadata, '_bind') is None:
            self._metadata.bind = self
        return rv

    async def close(self):
        if getattr(self._metadata, '_bind') is self:
            self._metadata.bind = None
        self._metadata = None
        return await super().close()

    def acquire(self, *, timeout=None, reuse=False, lazy=False):
        return GinoAcquireContext(self, timeout, reuse, lazy)

    async def _acquire(self, timeout, reuse=False, lazy=False):
        root = None
        local = get_local()
        if reuse and local:
            stack = local.get('connection_stack')
            if stack:
                root = getattr(stack[-1], '_root')
        conn = LazyConnection(self, timeout, root, self._connection_class,
                              self._loop)
        if local is not None:
            local.setdefault('connection_stack', deque()).append(conn)
        if not lazy:
            await conn.get_connection()
        return conn

    async def release(self, connection):
        if type(connection) is not LazyConnection or \
                getattr(connection, '_pool', None) is not self:
            raise InterfaceError(
                'Pool.release() received invalid connection: '
                '{connection!r} is not a member of this pool'.format(
                    connection=connection))

        ctx = get_local()
        if ctx is not None:
            stack = ctx.get('connection_stack')
            if not stack or stack[-1] is not connection:
                raise InterfaceError('Wrong release order')
            stack.pop()
            if not stack:
                ctx.pop('connection_stack')

        return await connection.release(close=True)

    @staticmethod
    def get_current_connection():
        local = get_local()
        if local:
            stack = local.get('connection_stack')
            if stack:
                return stack[-1]

    @property
    def execution_options(self):
        return self._execution_options

    @property
    def metadata(self):
        return self._metadata

    @property
    def dialect(self):
        return self._metadata.dialect

    async def all(self, clause, *multiparams, **params):
        return await self.dialect.do_all(self, clause,
                                         *multiparams, **params)

    async def first(self, clause, *multiparams, **params):
        return await self.dialect.do_first(self, clause,
                                           *multiparams, **params)

    async def scalar(self, clause, *multiparams, **params):
        return await self.dialect.do_scalar(self, clause,
                                            *multiparams, **params)

    async def status(self, clause, *multiparams, **params):
        return await self.dialect.do_status(self, clause,
                                            *multiparams, **params)



# Waiting for https://www.python.org/dev/peps/pep-0550/

import asyncio

_original_task_factories = {}


def get_local():
    return getattr(asyncio.Task.current_task(), 'task_local', None)


def task_factory(loop, coro):
    orig = _original_task_factories.get(loop)
    loop.set_task_factory(orig)
    task = loop.create_task(coro)
    loop.set_task_factory(task_factory)

    if getattr(task, '_source_traceback', None):
        del getattr(task, '_source_traceback')[-2:]

    task.task_local = get_local() or {}

    return task


def enable_task_local(loop=None):
    if loop is None:
        loop = asyncio.get_event_loop()
    if loop in _original_task_factories:
        return
    _original_task_factories[loop] = loop.get_task_factory()
    loop.set_task_factory(task_factory)


def disable_task_local(loop=None):
    if loop is None:
        loop = asyncio.get_event_loop()
    if loop in _original_task_factories:
        loop.set_task_factory(_original_task_factories.pop(loop))



import sys
import weakref

import sqlalchemy as sa
from sqlalchemy.sql.base import Executable
from sqlalchemy.dialects import postgresql as sa_pg

from .connection import GinoConnection
from .crud import CRUDModel
from .declarative import declarative_base
from .dialect import AsyncpgDialect, GinoCursorFactory
from .pool import GinoPool
from . import json_support


class GinoTransaction:
    __slots__ = ('_conn_ctx', '_isolation', '_readonly', '_deferrable', '_ctx')

    def __init__(self, conn_ctx, isolation, readonly, deferrable):
        self._conn_ctx = conn_ctx
        self._isolation = isolation
        self._readonly = readonly
        self._deferrable = deferrable
        self._ctx = None

    async def __aenter__(self):
        conn = await self._conn_ctx.__aenter__()
        self._ctx = conn.transaction(isolation=self._isolation,
                                     readonly=self._readonly,
                                     deferrable=self._deferrable)
        return conn, await self._ctx.__aenter__()

    async def __aexit__(self, extype, ex, tb):
        try:
            await self._ctx.__aexit__(extype, ex, tb)
        except:
            await self._conn_ctx.__aexit__(*sys.exc_info())
            raise
        else:
            await self._conn_ctx.__aexit__(extype, ex, tb)


class ConnectionAcquireContext:
    __slots__ = ('_connection',)

    def __init__(self, connection):
        self._connection = connection

    async def __aenter__(self):
        return self._connection

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    def __await__(self):
        return self

    def __next__(self):
        raise StopIteration(self._connection)


class GinoExecutor:
    __slots__ = ('_query',)

    def __init__(self, query):
        self._query = query

    @property
    def query(self):
        return self._query

    def model(self, model):
        self._query = self._query.execution_options(model=weakref.ref(model))
        return self

    def return_model(self, switch):
        self._query = self._query.execution_options(return_model=switch)
        return self

    def timeout(self, timeout):
        self._query = self._query.execution_options(timeout)
        return self

    def get_bind(self, bind):
        return bind or self._query.bind

    async def all(self, *multiparams, bind=None, **params):
        bind = self.get_bind(bind)
        return await bind.metadata.dialect.do_all(
            bind, self._query, *multiparams, **params)

    async def first(self, *multiparams, bind=None, **params):
        bind = self.get_bind(bind)
        return await bind.metadata.dialect.do_first(
            bind, self._query, *multiparams, **params)

    async def scalar(self, *multiparams, bind=None, **params):
        bind = self.get_bind(bind)
        return await bind.metadata.dialect.do_scalar(
            bind, self._query, *multiparams, **params)

    async def status(self, *multiparams, bind=None, **params):
        """
        You can parse the return value like this: https://git.io/v7oze
        """
        bind = self.get_bind(bind)
        return await bind.metadata.dialect.do_status(
            bind, self._query, *multiparams, **params)

    def iterate(self, *multiparams, connection=None, **params):
        def env_factory():
            conn = self.get_bind(connection)
            return conn, conn.metadata
        return GinoCursorFactory(env_factory, self._query, multiparams, params)


class Gino(sa.MetaData):
    default_model_classes = (CRUDModel,)

    def __init__(self, bind=None, dialect=None, model_classes=None,
                 query_ext=True, **kwargs):
        self._bind = None
        super().__init__(bind=bind, **kwargs)
        self.dialect = dialect or AsyncpgDialect()
        if model_classes is None:
            model_classes = self.default_model_classes
        self.Model = declarative_base(self, model_classes)
        for mod in sa, sa_pg, json_support:
            for key in mod.__all__:
                if not hasattr(self, key):
                    setattr(self, key, getattr(mod, key))
        if query_ext:
            Executable.gino = property(GinoExecutor)

    @property
    def bind(self):
        return getattr(self._bind, 'get_current_connection',
                       lambda: None)() or self._bind

    # noinspection PyMethodOverriding
    @bind.setter
    def bind(self, val):
        self._bind = val

    def create_pool(self, dsn=None, *,
                    min_size=10,
                    max_size=10,
                    max_queries=50000,
                    max_inactive_connection_lifetime=300.0,
                    setup=None,
                    init=None,
                    loop=None,
                    connection_class=GinoConnection,
                    **connect_kwargs):
        if not issubclass(connection_class, GinoConnection):
            raise TypeError(
                'connection_class is expected to be a subclass of '
                'gino.GinoConnection, got {!r}'.format(connection_class))

        pool = GinoPool(
            self, dsn,
            connection_class=connection_class,
            min_size=min_size, max_size=max_size,
            max_queries=max_queries, loop=loop, setup=setup, init=init,
            max_inactive_connection_lifetime=max_inactive_connection_lifetime,
            **connect_kwargs)
        return pool

    def compile(self, elem, *multiparams, **params):
        return self.dialect.compile(elem, *multiparams, **params)

    async def all(self, clause, *multiparams, bind=None, **params):
        return await self.dialect.do_all(
            bind or self.bind, clause, *multiparams, **params)

    async def first(self, clause, *multiparams, bind=None, **params):
        return await self.dialect.do_first(
            bind or self.bind, clause, *multiparams, **params)

    async def scalar(self, clause, *multiparams, bind=None, **params):
        return await self.dialect.do_scalar(
            bind or self.bind, clause, *multiparams, **params)

    async def status(self, clause, *multiparams, bind=None, **params):
        return await self.dialect.do_status(
            bind or self.bind, clause, *multiparams, **params)

    def iterate(self, clause, *multiparams, connection=None, **params):
        return GinoCursorFactory(lambda: (connection or self.bind, self),
                                 clause, multiparams, params)

    def acquire(self, *, timeout=None, reuse=True, lazy=False):
        method = getattr(self._bind, 'acquire', None)
        if method is None:
            return ConnectionAcquireContext(self._bind)
        else:
            return method(timeout=timeout, reuse=reuse, lazy=lazy)

    def transaction(self, *, isolation='read_committed', readonly=False,
                    deferrable=False, timeout=None, reuse=True):
        return GinoTransaction(self.acquire(timeout=timeout, reuse=reuse),
                               isolation, readonly, deferrable)



import random

import pytest
from .models import User, Friendship

pytestmark = pytest.mark.asyncio


async def test_create(asyncpg_pool):
    nickname = 'test_create_{}'.format(random.random())
    u = await User.create(bind=asyncpg_pool, nickname=nickname)
    assert u.id is not None
    assert u.nickname == nickname
    return u


async def test_get(asyncpg_pool):
    u1 = await test_create(asyncpg_pool)
    u2 = await User.get(u1.id, bind=asyncpg_pool)
    assert u1.id == u2.id
    assert u1.nickname == u2.nickname
    assert u1 is not u2


async def test_get_multiple_primary_key(asyncpg_pool):
    u1 = await test_create(asyncpg_pool)
    u2 = await test_create(asyncpg_pool)
    await Friendship.create(bind=asyncpg_pool, my_id=u1.id, friend_id=u2.id)
    f = await Friendship.get((u1.id, u2.id), bind=asyncpg_pool)
    assert f
    assert f.my_id == u1.id
    assert f.friend_id == u2.id


async def test_connection_as_bind(asyncpg_pool):
    async with asyncpg_pool.acquire() as conn:
        await test_get(conn)


async def test_update(asyncpg_pool, random_name):
    u1 = await test_create(asyncpg_pool)
    await u1.update(nickname=random_name).apply(bind=asyncpg_pool)
    u2 = await User.get(u1.id, bind=asyncpg_pool)
    assert u2.nickname == random_name


async def test_update_multiple_primary_key(asyncpg_pool):
    u1 = await test_create(asyncpg_pool)
    u2 = await test_create(asyncpg_pool)
    u3 = await test_create(asyncpg_pool)
    await Friendship.create(bind=asyncpg_pool, my_id=u1.id, friend_id=u2.id)
    f = await Friendship.get((u1.id, u2.id), bind=asyncpg_pool)
    await f.update(my_id=u2.id, friend_id=u3.id).apply(bind=asyncpg_pool)
    f2 = await Friendship.get((u2.id, u3.id), bind=asyncpg_pool)
    assert f2


async def test_delete(asyncpg_pool):
    u1 = await test_create(asyncpg_pool)
    await u1.delete(bind=asyncpg_pool)
    u2 = await User.get(u1.id, bind=asyncpg_pool)
    assert not u2


async def test_delete_multiple_primary_key(asyncpg_pool):
    u1 = await test_create(asyncpg_pool)
    u2 = await test_create(asyncpg_pool)
    f = await Friendship.create(bind=asyncpg_pool, my_id=u1.id, friend_id=u2.id)
    await f.delete(bind=asyncpg_pool)
    f2 = await Friendship.get((u1.id, u2.id), bind=asyncpg_pool)
    assert not f2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages, Extension
from pkg_resources import resource_filename

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()


def req_file(filename):
    with open(filename) as f:
        content = f.readlines()
    return [x.strip() for x in content if x.strip()]

setup_requirements = [
    'Cython>=0.24',
    'pytest-runner',
]

test_requirements = [
    'pytest',
    'pytest-asyncio',
    'psycopg2',
]


class LazyExtension(Extension):
    def __init__(self, *args, **kwargs):
        self._include_dirs = []
        super().__init__(*args, **kwargs)

    @property
    def include_dirs(self):
        return self._include_dirs + [resource_filename('asyncpg', 'protocol')]

    @include_dirs.setter
    def include_dirs(self, val):
        self._include_dirs = val


setup(
    name='gino',
    version='0.5.0',
    description="GINO Is Not ORM - "
                "a Python ORM on asyncpg and SQLAlchemy core.",
    long_description=readme + '\n\n' + history,
    author="Fantix King",
    author_email='fantix.king@gmail.com',
    url='https://github.com/fantix/gino',
    packages=find_packages(),
    ext_modules=[
        LazyExtension('gino.record', ['gino/record.pyx']),
    ],
    include_package_data=True,
    install_requires=req_file('requirements.txt'),
    license="BSD license",
    zip_safe=False,
    keywords='gino',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



"""
GINO provides a convenient plugin for integrating with Tornado_ webserver.
It consists of two parts, each of them is optional.

.. _Tornado: http://www.tornadoweb.org

.. warning::

    Tornado doesn't wrap request handlers to asyncio tasks, hence
    task locals doesn't work in request handlers by default. To fix this,
    you may either redefine ``_execute()`` method on you handlers to wrap
    request processing into a task, or simply use
    :py:class:`gino.ext.tornado.AsyncioRequestHandler`
    as a base class for all of your handlers.

    See `integrate GINO with application and request handlers`_ for more details.


Provide tornado-specific methods on models
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

GINO can provide a webframework-aware ``.get_or_404()`` method which work
similar to ``.get()`` but raises an appropriate error whenever requested object
not found. In case of tornado, an appropriate error is
``tornado.web.HTTPError(404)``.

To have it working, simply use :py:class:`gino.ext.tornado.Gino` as your
database metadata.


Integrate GINO with application and request handlers
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

In addition to ``.get_or_404()``, GINO provides bases for application and
request handler objects.

Inherit your application class from :py:class:`gino.ext.tornado.Application`
to automate connection pool management and provide access to the database
object to all of your request handlers via ``self.application.db``.

Inherit your request handlers from
:py:class:`gino.ext.tornado.AsyncioRequestHandler` to enable task locals
support.

Inherit your request handlers from :py:class:`gino.ext.tornado.GinoRequestHandler`
to enable active connection management.
Note that :py:class:`gino.ext.tornado.GinoRequestHandler` requires your application
to have a ``db`` property with ``acquire`` coroutine so its best to use it
with :py:class:`gino.ext.tornado.Application`.


Settings defined by this extension
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

GINO would define some options for database configuration. Use them with
the standard ``tornado.options`` module:

- ``'db_host'`` -- if not set, ``localhost``;
- ``'db_port'`` -- if not set, ``5432``;
- ``'db_user'`` -- if not set, ``postgres``;
- ``'db_password'`` -- if not set, empty string;
- ``'db_database'`` -- if not set, ``postgres``;
- ``'db_pool_min_size'`` -- number of connection the pool will be initialized
  with. Default is ``5``;
- ``'db_pool_max_size'`` -- max number of connections in the pool.
  Default is ``10``;
- ``'db_pool_max_inactive_conn_lifetime'`` -- number of seconds after which
  inactive connections in the pool will be closed.  Pass ``0`` to disable this
  mechanism. Default is ``300``;
- ``'db_pool_max_queries '`` -- number of queries after a connection is closed
  and replaced with a new connection. Default is ``50000``.


An example application
^^^^^^^^^^^^^^^^^^^^^^

A helloworld application that uses tornado and GINO may look like this:

.. code-block:: python

    import tornado.web
    import tornado.ioloop
    import tornado.options
    import tornado.escape

    from gino.ext.tornado import Gino, Application, GinoRequestHandler


    # Define your database metadata
    # -----------------------------

    db = Gino()


    # Define tables as you would normally do
    # --------------------------------------

    class User(db.Model):
        __tablename__ = 'users'

        id = db.Column(db.Integer(), primary_key=True, autoincrement=True)
        nickname = db.Column(db.Unicode(), nullable=False)


    # Now just use your tables
    # ------------------------

    class AllUsers(GinoRequestHandler):
        async def get(self):
            users = await User.query.gino.all()

            for user in users:
                url = self.application.reverse_url('user', user.id)
                nickname = tornado.escape.xhtml_escape(user.nickname)
                self.write(f'<a href="{url}">{nickname}</a><br/>')


    class GetUser(GinoRequestHandler):
        async def get(self, uid):
            user: User = await User.get_or_404(int(uid))
            self.write(f'Hi, {user.nickname}!')


    if __name__ == '__main__':
        tornado.options.parse_command_line()
        tornado.ioloop.IOLoop.configure('tornado.platform.asyncio.AsyncIOMainLoop')

        app = Application([
            tornado.web.URLSpec(r'/', AllUsers, name='index'),
            tornado.web.URLSpec(r'/user/(?P<uid>[0-9]+)', GetUser, name='user')
        ], debug=True)

        loop = tornado.ioloop.IOLoop.current().asyncio_loop

        # If you intend to use HTTPServer in multiprocessed environment,
        # call the app.late_init method after calling HTTPServer.start(n).
        # This will create one connection pool per process.
        loop.run_until_complete(app.late_init(db))

        app.listen(8888)

        loop.run_forever()

API reference
^^^^^^^^^^^^^

"""

import asyncio
import sys
import typing

import tornado.ioloop
import tornado.iostream
import tornado.log
import tornado.platform.asyncio
import tornado.web

from tornado.options import options as _options, define as _define

from ..api import Gino as _Gino
from ..local import enable_task_local as _enable_task_local


def _assert_not_negative(name):
    def inner(value):
        if value < 0:
            raise ValueError(f'{name} should be non-negative')
    return inner


_define('db_host', 'localhost', str, group='database')
_define('db_port', 5432, int, group='database')
_define('db_user', 'postgres', str, group='database')
_define('db_password', 'password', str, group='database')
_define('db_database', 'postgres', str, group='database')
_define('db_pool_min_size', 5, int, group='database',
        callback=_assert_not_negative('db_pool_min_size'))
_define('db_pool_max_size', 10, int, group='database',
        callback=_assert_not_negative('db_pool_max_size'))
_define('db_pool_max_inactive_conn_lifetime', 300, float, group='database',
        callback=_assert_not_negative('db_pool_max_inactive_conn_lifetime'))
_define('db_pool_max_queries', 50000, int, group='database',
        callback=_assert_not_negative('db_pool_max_queries'))


class TornadoModelMixin:
    @classmethod
    async def get_or_404(cls, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        rv = await cls.get(*args, **kwargs)
        if rv is None:
            raise tornado.web.HTTPError(404)
        return rv


class Gino(_Gino):
    """
    Base class for GINO database.

    Using this class as a metadata for your database adds an additional
    ``get_or_404()`` method to all of your table classes.

    """

    default_model_classes = _Gino.default_model_classes + (TornadoModelMixin,)

    if typing.TYPE_CHECKING:
        # Typehints to enable autocompletion on all Gino.Model-derived classes

        from ..crud import CRUDModel as __CRUDModel
        from ..declarative import ModelType as __ModelType

        class Model(__CRUDModel, TornadoModelMixin, metaclass=__ModelType):
            ...


class Application(tornado.web.Application):
    """
    Base application that provides access to the database object and defines
    a convenient method for initializing all the database-related stuff.

    """

    #: The database object associated with this application.
    #: Use :py:meth:`~.late_init()` to init this or set it manually.
    db: Gino = None

    #: If ``True``, enables ``GinoRequestHandler`` to create lazy connections.
    #:
    #: See :py:attr:`~.GinoRequestHandler.use_connection_for_request`
    #: for more info.
    use_connection_for_request: bool = True

    async def late_init(self, db: Gino, *, loop=None, options=_options):
        """
        Initialize this application with a database object.

        This method does a few things to setup application for working with
        the database:

        - it enables task local storage;
        - creates a connection pool and binds it to the passed database object;
        - populates :py:attr:`~.db`.

        :param db: the :py:class:`gino.ext.tornado.Gino()` class instance that
            will be used in this application.
        :param loop: io loop that will be used to run heep server, either
            tornado's or asyncio's.
        :param options: a tornado's ``OptionParser()`` instance or any
            dictionary-like object with the database settings. Default is to
            use ``tornado.options.options`` global.

        """

        if loop is None:
            loop = tornado.ioloop.IOLoop.current()
        if isinstance(loop, tornado.platform.asyncio.BaseAsyncIOLoop):
            asyncio_loop = loop.asyncio_loop
        elif isinstance(loop, asyncio.BaseEventLoop):
            asyncio_loop = loop
        else:
            raise RuntimeError('AsyncIOLoop is required to run GINO')

        _enable_task_local(asyncio_loop)

        self.db: Gino = db
        await db.create_pool(
            host=options['db_host'],
            port=options['db_port'],
            user=options['db_user'],
            password=options['db_password'],
            database=options['db_database'],
            min_size=options['db_pool_min_size'],
            max_size=options['db_pool_max_size'],
            max_inactive_connection_lifetime=(
                options['db_pool_max_inactive_conn_lifetime']
            ),
            max_queries=options['db_pool_max_queries'],
            loop=asyncio_loop
        )


# noinspection PyAbstractClass
class AsyncioRequestHandler(tornado.web.RequestHandler):
    """
    This class enables support for task locals by wrapping the ``_execute()``
    method into ``asyncio.Task`` instances.

    """

    application: tornado.web.Application

    async def _setup_connection(self):
        """
        Hook for creating connection.

        """
        pass

    async def _teardown_connection(self):
        """
        Hook for destroying connection.

        """
        pass

    def _execute(self, transforms, *args, **kwargs):
        loop = tornado.ioloop.IOLoop.current()

        if not isinstance(loop, tornado.platform.asyncio.BaseAsyncIOLoop):
            raise RuntimeError('AsyncIOLoop is required to run GINO')

        asyncio_loop = loop.asyncio_loop

        return asyncio.ensure_future(
            self._do_execute(transforms, *args, **kwargs),
            loop=asyncio_loop
        )

    async def _do_execute(self, transforms, *args, **kwargs):
        """
        An actual asyncio-compatible implementation on the ``_execute``.

        This function just takes the original generator ``_execute.__wrapped__``
        and manages to pass its futures to the underlying asyncio loop.

        It also calls ``_setup_connection`` and ``_teardown_connection``
        methods and manages all errors that happen there.

        """

        self._transforms = transforms

        try:
            gen = super()._execute.__wrapped__(
                self, transforms, *args, **kwargs
            )

            data = None
            exc_info = None

            await self._setup_connection()

            try:
                while True:
                    try:
                        if exc_info is None:
                            coro = gen.send(data)
                        else:
                            coro = gen.throw(*exc_info)
                    except StopIteration:
                        return
                    exc_info = None
                    data = None
                    # noinspection PyBroadException
                    try:
                        data = await coro
                    except:
                        exc_info = sys.exc_info()
            finally:
                await self._teardown_connection()
        except Exception as e:
            # noinspection PyBroadException
            try:
                self._handle_request_exception(e)
            except Exception:
                tornado.log.app_log.exception("exception in exception handler")
        finally:
            if self._prepared_future is not None:
                if not self._prepared_future.done():
                    self._prepared_future.set_result(None)


# noinspection PyAbstractClass
class GinoRequestHandler(AsyncioRequestHandler):
    """
    Base class for all request handlers that use GINO.

    In addition to features provided by :py:class:`~.AsyncioRequestHandler`,
    this class manages lazy connections for each request.

    """

    application: Application

    __db_connection = None

    @property
    def use_connection_for_request(self):
        """
        If ``True``, a lazy connection is created for each request.

        That is, whenever the first query occurs, a new connection is
        borrowed from the application's db object. All succeeding queries made
        within this request will reuse that connection. The connection will be
        returned to the pool once the request is finished or the
        :py:meth:`~.release_connection()` method is called explicitly.

        This property is equal to
        :py:attr:`Application.use_connection_for_request`
        by default.

        """
        return self.application.use_connection_for_request

    @property
    def db(self):
        """
        Access to the database object.

        This property is equal to :py:attr:`Application.db` by default.

        """
        return self.application.db

    async def _setup_connection(self):
        if self.use_connection_for_request:
            self.__db_connection = await self.db.acquire(lazy=True)

    async def _teardown_connection(self):
        if self.__db_connection is not None:
            await self.__db_connection.release()
            self.__db_connection = None

    @property
    def db_connection(self):
        """
        The actual connection associated with this request or ``None`` if
        ``use_connection_for_request`` is ``False``.

        """
        return self.__db_connection

    async def release_connection(self):
        """
        Return the connection associated with this request back to the pool.

        """
        await self._teardown_connection()



from asyncpg.connection import Connection

from .dialect import GinoCursorFactory


class GinoConnection(Connection):
    __slots__ = ('_metadata', '_execution_options')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata = None
        self._execution_options = {}

    def _set_metadata(self, metadata):
        self._metadata = metadata

    @property
    def metadata(self):
        return self._metadata

    @property
    def execution_options(self):
        return self._execution_options

    @property
    def dialect(self):
        return self._metadata.dialect

    async def all(self, clause, *multiparams, **params):
        return await self.dialect.do_all(self, clause, *multiparams, **params)

    async def first(self, clause, *multiparams, **params):
        return await self.dialect.do_first(self, clause,
                                           *multiparams, **params)

    async def scalar(self, clause, *multiparams, **params):
        return await self.dialect.do_scalar(self, clause,
                                            *multiparams, **params)

    async def status(self, clause, *multiparams, **params):
        return await self.dialect.do_status(self, clause,
                                            *multiparams, **params)

    def iterate(self, clause, *multiparams, **params):
        return GinoCursorFactory(lambda: (self, self.metadata), clause,
                                 multiparams, params)



import itertools
import weakref

import sqlalchemy as sa
from sqlalchemy.sql import ClauseElement
from sqlalchemy.dialects import postgresql as sa_pg

from . import json_support
from .declarative import Model
from .exceptions import NoSuchRowError

DEFAULT = object()


class Query:
    def __get__(self, instance, owner):
        q = sa.select([owner.__table__])
        if instance is not None:
            q = instance.append_where_primary_key(q)
        return q.execution_options(model=weakref.ref(owner))


class Select:
    def __get__(self, instance, owner):
        def select(*args):
            q = sa.select([getattr(owner, x) for x in args])
            if instance is not None:
                q = instance.append_where_primary_key(q)
            return q.execution_options(model=weakref.ref(owner),
                                       return_model=False)
        return select


class Update:
    def __get__(self, instance, owner):
        if instance is None:
            q = owner.__table__.update()
            return q.execution_options(model=weakref.ref(owner))
        else:
            # noinspection PyProtectedMember
            return instance._update


class Delete:
    def __get__(self, instance, owner):
        if instance is None:
            q = owner.__table__.delete()
            return q.execution_options(model=weakref.ref(owner))
        else:
            # noinspection PyProtectedMember
            return instance._delete


class UpdateRequest:
    def __init__(self, instance):
        self._instance = instance
        self._values = {}
        self._props = {}
        self._literal = True
        self._clause = self._instance.append_where_primary_key(
            type(self._instance).update)

    def set(self, key, value):
        self._values[key] = value

    def set_prop(self, prop, value):
        if isinstance(value, ClauseElement):
            self._literal = False
        self._props[prop] = value

    async def apply(self, bind=None, timeout=DEFAULT):
        cls = type(self._instance)
        values = self._values.copy()

        # handle JSON columns
        json_updates = {}
        for prop, value in self._props.items():
            value = prop.save(self._instance, value)
            updates = json_updates.setdefault(prop.column_name, {})
            if self._literal:
                updates[prop.name] = value
            else:
                if isinstance(value, int):
                    value = sa.cast(value, sa.BigInteger)
                elif not isinstance(value, ClauseElement):
                    value = sa.cast(value, sa.Unicode)
                updates[sa.cast(prop.name, sa.Unicode)] = value
        for column_name, updates in json_updates.items():
            column = getattr(cls, column_name)
            if self._literal:
                values[column_name] = column.concat(updates)
            else:
                if isinstance(column.type, sa_pg.JSONB):
                    func = sa.func.jsonb_build_object
                else:
                    func = sa.func.json_build_object
                values[column_name] = column.concat(
                    func(*itertools.chain(*updates.items())))

        opts = dict(return_model=False)
        if timeout is not DEFAULT:
            opts['timeout'] = timeout
        clause = self._clause.values(
            **values,
        ).returning(
            *[getattr(cls, key) for key in values],
        ).execution_options(**opts)
        row = await cls.__metadata__.first(clause, bind=bind)
        if not row:
            raise NoSuchRowError()
        self._instance.__values__.update(row)
        for prop in self._props:
            prop.reload(self._instance)
        return self


class CRUDModel(Model):
    query = Query()
    update = Update()
    delete = Delete()
    select = Select()

    def __init__(self, **values):
        super().__init__()
        self.__profile__ = None
        # noinspection PyCallingNonCallable
        self.update(**values)

    @classmethod
    def _init_table(cls, sub_cls):
        rv = Model._init_table(sub_cls)
        if rv is not None:
            rv.__model__ = weakref.ref(sub_cls)
        return rv

    @classmethod
    async def create(cls, bind=None, timeout=DEFAULT, **values):
        rv = cls(**values)

        # handle JSON properties
        props = []
        for key, value in values.items():
            prop = cls.__dict__.get(key)
            if isinstance(prop, json_support.JSONProperty):
                prop.save(rv)
                props.append(prop)
        for key, prop in cls.__dict__.items():
            if key in values:
                continue
            if isinstance(prop, json_support.JSONProperty):
                if prop.default is None or prop.after_get.method is not None:
                    continue
                setattr(rv, key, getattr(rv, key))
                prop.save(rv)
                props.append(prop)

        opts = dict(return_model=False, model=cls)
        if timeout is not DEFAULT:
            opts['timeout'] = timeout
        q = cls.__table__.insert().values(**rv.__values__).returning(
            sa.text('*')).execution_options(**opts)
        row = await cls.__metadata__.first(q, bind=bind)
        rv.__values__.update(row)
        rv.__profile__ = None
        return rv

    @classmethod
    async def get(cls, ident, bind=None, timeout=DEFAULT):
        if hasattr(ident, '__iter__'):
            ident_ = list(ident)
        else:
            ident_ = [ident]
        columns = cls.__table__.primary_key.columns
        if len(ident_) != len(columns):
            raise ValueError(
                'Incorrect number of values as primary key: '
                'expected {}, got {}.'.format(
                    len(columns), len(ident_)))
        clause = cls.query
        for i, c in enumerate(columns):
            clause = clause.where(c == ident_[i])
        if timeout is not DEFAULT:
            clause = clause.execution_options(timeout=timeout)
        return await cls.__metadata__.first(clause, bind=bind)

    def append_where_primary_key(self, q):
        for c in self.__table__.primary_key.columns:
            q = q.where(c == getattr(self, c.name))
        return q

    def _update(self, **values):
        cls = type(self)
        rv = UpdateRequest(self)
        for key, value in values.items():
            prop = cls.__dict__.get(key)
            if isinstance(prop, json_support.JSONProperty):
                value_from = '__profile__'
                method = rv.set_prop
                k = prop
            else:
                value_from = '__values__'
                method = rv.set
                k = key
            if not isinstance(value, ClauseElement):
                setattr(self, key, value)
                value = getattr(self, value_from)[key]
            method(k, value)
        return rv

    async def _delete(self, bind=None, timeout=DEFAULT):
        cls = type(self)
        clause = self.append_where_primary_key(cls.delete)
        if timeout is not DEFAULT:
            clause = clause.execution_options(timeout=timeout)
        return await self.__metadata__.status(clause, bind=bind)

    def to_dict(self):
        cls = type(self)
        keys = set(c.name for c in cls)
        for key, prop in cls.__dict__.items():
            if isinstance(prop, json_support.JSONProperty):
                keys.add(key)
                keys.discard(prop.column_name)
        return dict((k, getattr(self, k)) for k in keys)



import os

from gino import Gino

DB_ARGS = dict(
    host=os.getenv('DB_HOST', 'localhost'),
    port=os.getenv('DB_PORT', 5432),
    user=os.getenv('DB_USER', 'postgres'),
    password=os.getenv('DB_PASS', ''),
    database=os.getenv('DB_NAME', 'postgres'),
)
db = Gino()


class User(db.Model):
    __tablename__ = 'gino_users'

    id = db.Column(db.BigInteger(), primary_key=True)
    nickname = db.Column(db.Unicode(), default='noname')

    def __repr__(self):
        return '{}<{}>'.format(self.nickname, self.id)


class Friendship(db.Model):
    __tablename__ = 'gino_friendship'

    my_id = db.Column(db.BigInteger(), primary_key=True)
    friend_id = db.Column(db.BigInteger(), primary_key=True)

    def __repr__(self):
        return 'Friends<{}, {}>'.format(self.my_id, self.friend_id)



import weakref

from sqlalchemy import cutils, util
from sqlalchemy.dialects.postgresql.base import (
    PGCompiler,
    PGDialect,
    PGExecutionContext,
)

from .record import update_record

DEFAULT = object()


class AsyncpgCompiler(PGCompiler):
    @property
    def bindtemplate(self):
        return self._bindtemplate

    @bindtemplate.setter
    def bindtemplate(self, val):
        # noinspection PyAttributeOutsideInit
        self._bindtemplate = val.replace(':', '$')


class NoopConnection:
    def __init__(self, dialect, execution_options):
        self.dialect = dialect
        self._execution_options = execution_options or {}

    def cursor(self):
        pass


# noinspection PyAbstractClass
class AsyncpgExecutionContext(PGExecutionContext):
    @classmethod
    def init_clause(cls, dialect, elem, multiparams, params,
                    execution_options):
        # partially copied from:
        # sqlalchemy.engine.base.Connection:_execute_clauseelement
        # noinspection PyProtectedMember
        distilled_params = cutils._distill_params(multiparams, params)
        if distilled_params:
            # note this is usually dict but we support RowProxy
            # as well; but dict.keys() as an iterable is OK
            keys = distilled_params[0].keys()
        else:
            keys = []
        compiled_sql = elem.compile(
            dialect=dialect, column_keys=keys,
            inline=len(distilled_params) > 1,
        )
        if execution_options:
            execution_options = dict(execution_options)
        else:
            execution_options = {}
        for opt in ('return_model', 'model', 'timeout'):
            if opt in compiled_sql.execution_options:
                execution_options.pop(opt, None)
        conn = NoopConnection(dialect, execution_options)
        rv = cls._init_compiled(
            dialect, conn, conn, compiled_sql, distilled_params)
        return rv

    @util.memoized_property
    def return_model(self):
        # noinspection PyUnresolvedReferences
        return self.execution_options.get('return_model', True)

    @util.memoized_property
    def model(self):
        # noinspection PyUnresolvedReferences
        rv = self.execution_options.get('model', None)
        if isinstance(rv, weakref.ref):
            rv = rv()
        return rv

    @util.memoized_property
    def timeout(self):
        # noinspection PyUnresolvedReferences
        return self.execution_options.get('timeout', None)

    def from_row(self, row, return_row=False):
        if self.model is None or row is None:
            return row
        for index, (key, value) in enumerate(row.items()):
            type_ = getattr(getattr(self.model, key), 'type', None)
            if type_ is not None:
                processor = self.get_result_processor(type_, None, None)
                if processor:
                    new_value = processor(value)
                    if new_value is not value:
                        update_record(row, index, new_value)
        if not return_row and self.return_model:
            rv = self.model()
            rv.__values__.update(row)
        else:
            rv = row
        return rv


class GinoCursorFactory:
    def __init__(self, env_factory, clause, multiparams, params):
        self._env_factory = env_factory
        self._context = None
        self._timeout = None
        self._clause = clause
        self._multiparams = multiparams
        self._params = params

    @property
    def timeout(self):
        return self._timeout

    async def get_cursor_factory(self):
        connection, metadata = self._env_factory()
        self._context = metadata.dialect.execution_ctx_cls.init_clause(
            metadata.dialect, self._clause, self._multiparams, self._params,
            getattr(connection, 'execution_options', None))
        self._timeout = self._context.timeout
        return connection.cursor(self._context.statement,
                                 *self._context.parameters[0],
                                 timeout=self._timeout)

    @property
    def context(self):
        return self._context

    def __aiter__(self):
        return GinoCursorIterator(self)

    def __await__(self):
        return GinoCursor(self).async_init().__await__()


class GinoCursorIterator:
    def __init__(self, factory):
        self._factory = factory
        self._iterator = None

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self._iterator is None:
            factory = await self._factory.get_cursor_factory()
            self._iterator = factory.__aiter__()
        row = await self._iterator.__anext__()
        return self._factory.context.from_row(row)


class GinoCursor:
    def __init__(self, factory):
        self._factory = factory
        self._cursor = None

    async def async_init(self):
        factory = await self._factory.get_cursor_factory()
        self._cursor = await factory
        return self

    async def many(self, n, *, timeout=DEFAULT):
        if timeout is DEFAULT:
            timeout = self._factory.timeout
        rows = await self._cursor.fetch(n, timeout=timeout)
        return list(map(self._factory.context.from_row, rows))

    async def next(self, *, timeout=DEFAULT):
        if timeout is DEFAULT:
            timeout = self._factory.timeout
        row = await self._cursor.fetchrow(timeout=timeout)
        return self._factory.context.from_row(row)

    def __getattr__(self, item):
        return getattr(self._cursor, item)


# noinspection PyAbstractClass
class AsyncpgDialect(PGDialect):
    default_paramstyle = 'numeric'
    statement_compiler = AsyncpgCompiler
    execution_ctx_cls = AsyncpgExecutionContext

    def compile(self, elem, *multiparams, **params):
        context = self.execution_ctx_cls.init_clause(
            self, elem, multiparams, params, None)
        return context.statement, context.parameters[0]

    async def do_all(self, bind, clause, *multiparams, **params):
        context = self.execution_ctx_cls.init_clause(
            self, clause, multiparams, params,
            getattr(bind, 'execution_options', None))
        rows = await bind.fetch(context.statement, *context.parameters[0],
                                timeout=context.timeout)
        return list(map(context.from_row, rows))

    async def do_first(self, bind, clause, *multiparams, **params):
        context = self.execution_ctx_cls.init_clause(
            self, clause, multiparams, params,
            getattr(bind, 'execution_options', None))
        row = await bind.fetchrow(context.statement, *context.parameters[0],
                                  timeout=context.timeout)
        return context.from_row(row)

    async def do_scalar(self, bind, clause, *multiparams, **params):
        context = self.execution_ctx_cls.init_clause(
            self, clause, multiparams, params,
            getattr(bind, 'execution_options', None))
        row = await bind.fetchrow(context.statement, *context.parameters[0],
                                  timeout=context.timeout)
        if not row:
            return None
        return next(iter(context.from_row(row, return_row=True).values()))

    async def do_status(self, bind, clause, *multiparams, **params):
        context = self.execution_ctx_cls.init_clause(
            self, clause, multiparams, params,
            getattr(bind, 'execution_options', None))
        return await bind.execute(context.statement, *context.parameters[0],
                                  timeout=context.timeout)
